/**
 | File: bernoulli.c
 |
 | A procedure that returns am integer random number distributed
 | according to the Bernoulli distribution of 'n' trials with a
 | probability 'p'; uses 'rannum.c' to obtain random numbers with
 | uniform distribution.
 |
 | From: Press, Teukolsky, Vetterling and Flannery -- Numerical
 |       Recipes in C (2nd Ed.) -- Cambridge University Press
 |       (1992) -- Pagg. 294-295 .
 |
 | $Id: bernoulli.c,v 1.1 1999/05/18 13:37:46 loreti Exp $
 |
**/

#include <math.h>
#include "bernoulli.h"

unsigned int bernoulli(
  const unsigned int n,
  const double pp
) {
  unsigned int nl;
  double am, p;
  static double pc, plog, pclog, en, oldg;
  static double pold = -1.0;
  static unsigned int nold = 0;

/**
 | The binomial distribution is invariant changing 'pp' to its complement
 | to one, and the number of successfull tries to the complement to 'n';
 | 'am' is the mean of the deviate to be produced.
**/

  p = (pp <= 0.5 ? pp : 1.0 - pp);
  am = n * p;

/**
 | If 'n' is small, use the direct method.
**/

  if (n < 25) {
    unsigned int j;

    for (nl = 0,  j = 0;   j < n;   j++) {
      if (ranNum() < p) ++nl;
    }

/**
 | If 'n' is large and the probability small, approximates with a
 | Poisson deviate; if not, use the rejection method.
**/

  } else if (am < 1.0) {
    unsigned int j;
    double g, t;

    g = exp(-am);
    t = 1.0;
    for (j = 0;   j <= n;   j++) {
      t *= ranNum();
      if (t < g) break;
    }
    nl = (j > n ? n : j);

  } else {
    double sq, t;

/**
 | If 'n' and/or 'p' have changed, computes some useful numbers to be
 | saved for the next calls with the same values.
**/

    if (n != nold) {
      en = (nold = n);
      oldg = gammln(en + 1.0);
    }

    if (p != pold) {
      pc = 1.0 - (pold = p);
      plog = log(p);
      pclog = log(pc);
    }

    sq = sqrt(2.0 * am * pc);

    do {
      double em, y;

      do {
        double v1, v2;

/**
 | The following lines are equivalent to y = tan(\pi * ranNum())
**/

        do {
          v1 = 2.0 * ranNum() - 1.0;
          v2 = 2.0 * ranNum() - 1.0;
        } while (v1 * v1 + v2 * v2 > 1.0);
        y = v2 / v1;

        em = sq * y + am;
      } while (em < 0.0   ||   em >= (en + 1.0));
      nl = em;
      em = nl;                  /* Equivalent to em=floor(em) */
      t = 1.2 * sq * (1.0 + y * y) *
        exp(oldg - gammln(em + 1.0) - gammln(en - em + 1.0) + em * plog +
            (en - em) * pclog);
    } while (ranNum() > t);     /* Rejection method */
  }

  return (p == pp ? nl : n - nl);
}

