/**
 | File: boxmul.c
 |
 | An implementation of the Box-Muller method to obtain normally
 | distributed random numbers.  Uses 'rannum.c' to obtain random
 | numbers with uniform distribution.
 |
 | From: Press, Teukolsky, Vetterling and Flannery -- Numerical
 |       Recipes in C (2nd Ed.) -- Cambridge University Press
 |       (1992) -- Pagg. 288-290 .
 |
 | $Id: boxmul.c,v 1.1 1999/05/18 13:37:46 loreti Exp $
 |
**/

#include <math.h>
#include "boxmul.h"

double boxmul(void)
{
  static int    flag = 0;
  static double gset;
  double        fac, rsq, v1, v2;

  if (flag) {

/**
 | A valid number has been already computed and stored in gset;
 | just return it (and unset the flag).
**/

    flag = 0;
    return gset;

  } else {

/**
 | Pick two uniform random numbers v1 and v2 in the range [-1.0,+1.0] ,
 | in such a way that (v1,v2) is inside the unit circle; then applies
 | the Box-Muller transformation to get two normal random numbers from
 | them.  One is returned, the other saved for the next call.
**/
    do {
      v1  = 2.0 * ranNum() - 1.0;
      v2  = 2.0 * ranNum() - 1.0;
      rsq = v1 * v1 + v2 * v2;
    } while (rsq >= 1.0   ||   rsq == 0.0);

    fac  = sqrt(-2.0 * log(rsq) / rsq);
    gset = v1 * fac;
    flag = 1;

    return v2 * fac;
  }
}
