/**
 | File: poisson.c
 |
 | A procedure that returns an integer random number distributed
 | according to the Poisson distribution with mean 'xm'; uses
 | 'rannum.c' to obtain random numbers with uniform distribution.
 |
 | From: Press, Teukolsky, Vetterling and Flannery -- Numerical
 |       Recipes in C (2nd Ed.) -- Cambridge University Press
 |       (1992) -- Pagg. 294-295 .
 |
 | $Id: poisson.c,v 1.1 1999/05/18 13:37:46 loreti Exp $
 |
**/

#include <math.h>
#include "poisson.h"

unsigned int poisson(
  const double xm
) {
  double t;
  unsigned int m;
  static double alxm, g, sq;
  static double oldm = -1.0;

/**
 | If 'xm' is small, use the direct method; otherwise use the rejection
 | method.  In both cases, check if 'xm' has changed since last call:
 | if so, computes some useful numbers for the next calls with the same
 | 'xm'.
**/

  if (xm < 12.0) {
    if (xm != oldm) {
      g = exp(- (oldm = xm));
    }

/**
 | Instead of adding exponential deviates, it is equivalent to multiply
 | uniform deviates.  We never actually have to take the log; but merely
 | compare the result with the pre-computed exponential.
**/

    m = -1;
    t = 1.0;
    do {
      ++m;
      t *= ranNum();
    } while (t > g);

  } else {
    double em;

    if (xm != oldm) {
      sq = sqrt(2.0 * (oldm = xm));
      g = xm * (alxm = log(xm)) - gammln(xm + 1.0);
    }

    do {
      double y;

      do {
        double v1, v2;

/**
 | The following lines are equivalent to y = tan(\pi * ranNum())
**/

        do {
          v1 = 2.0 * ranNum() - 1.0;
          v2 = 2.0 * ranNum() - 1.0;
        } while (v1 * v1 + v2 * v2 > 1.0);
        y = v2 / v1;

        em = sq * y + xm;
      } while (em < 0.0);       /* Always reject in the forbidden region */

      m = em;
      em = m;                   /* Equivalent to em=floor(xm) */
      t = 0.9 * (1.0 + y * y) * exp(em * alxm - gammln(em + 1.0) - g);
    } while (ranNum() > t);     /* Rejection method */
  }

  return m;
}
