/**
 | File: rangamma.c
 |
 | A procedure that returns a random number distributed according
 | to the gamma distribution of integer order 'ia' (i.e. the waiting
 | time to the ia-th event in a Poisson process of unit mean); uses
 | 'rannum.c' to obtain random numbers with uniform distribution.
 |
 | From: Press, Teukolsky, Vetterling and Flannery -- Numerical
 |       Recipes in C (2nd Ed.) -- Cambridge University Press
 |       (1992) -- Pagg. 292-293 .
 |
 | $Id: rangamma.c,v 1.1 1999/05/18 13:37:46 loreti Exp $
 |
**/

#include <stdio.h>
#include <math.h>
#include "rangamma.h"

double rangamma(
  const unsigned int ia
) {
  double x;

  if (ia < 1) {
    fprintf(stderr, "rangamma: invalid ia = %u\n", ia);
    return 0;
  }

/**
 | For small values of 'ia', directly add waiting times; otherwise,
 | use the rejection method.
**/

  if (ia < 6) {
    double x;
    unsigned int j;

    for (j = 0,  x = 1.0;   j < ia;   j++) {
      x *= ranNum();
    }
    return (-log(x));
  } else {
    double e;

    do {
      double am, s, y;

      do {
        double v1, v2;

/**
 | The following lines are equivalent to y = tan(\pi * ranNum())
**/

        do {
          v1 = 2.0 * ranNum() - 1.0;
          v2 = 2.0 * ranNum() - 1.0;
        } while (v1 * v1 + v2 * v2 > 1.0);
        y = v2 / v1;

        am = ia - 1;
        s = sqrt(2.0 * am + 1.0);
        x = s * y + am;
      } while (x <= 0.0);       /* Always reject in the forbidden region */

      e = (1.0 + y * y) * exp(am * log(x / am) - s * y);
    } while (ranNum() > e);     /* Rejection method */

    return x;
  } /* if (ia < 6) */
}
