/*------------------------------------------------------*
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@padova.infn.it                        |
 | WWW:    http://wwwcdf.pd.infn.it/~loreti/mlo.html    |
 *------------------------------------------------------*

 See rannum.h for details about this procedure

 $Id: rannum.c,v 1.1 1999/05/18 13:37:46 loreti Exp $

 *------------------------------------------------------*/

#include <stdlib.h>             /* size_t */
#include "rannum.h"

#define P                  179
#define PM1            (P -  1)
#define Q              (P - 10)

#define STATE_SIZE          97
#define MANTISSA_SIZE       24

#define RANDOM_REALS  16777216.0
#define INIT_C          362436.0
#define INIT_CD        7654321.0
#define INIT_CM       16777213.0

static size_t  ni;
static size_t  nj;
static double  u[STATE_SIZE];
static double  c, cd, cm;

static unsigned int collapse(int, unsigned int);

static unsigned int collapse(
  int           anyInt,
  unsigned int  size
) {

/**
 | Return a value between 0 and size-1 inclusive; this value will be
 | anyInt itself if possible, otherwise another value in the required
 | interval.
**/

  if (anyInt < 0)           anyInt  = -(anyInt / 2);
  while (anyInt >= size)    anyInt /= 2;
  return                    anyInt;
}


void ranSetSeed(
  int seed_a,
  int seed_b
) {

/**
 | This procedure initialises the state table "u" for a lagged
 | Fibonacci sequence generator, filling it with random bits from
 | a small multiplicative congruential sequence; the auxilliary
 | variables "c", "ni" and "nj" are also initialized.  The seeds
 | are transformed into initial values in such a way that identical
 | results are guaranteed across a wide variety of machines.
**/

  double        s, bit;
  unsigned int  ii, jj, kk, mm;
  unsigned int  ll;
  unsigned int  sd;
  unsigned int  elt, bit_number;

  sd = collapse(seed_a, PM1 * PM1);
  ii = 1 + sd / PM1;
  jj = 1 + sd % PM1;
  sd = collapse(seed_b, PM1 * Q);
  kk = 1 + sd / PM1;
  ll = sd % Q;

  if (ii == 1   &&   jj == 1   &&   kk == 1)  ii = 2;

  ni  = STATE_SIZE - 1;
  nj  = STATE_SIZE / 3;
  c   = INIT_C;
  c  /= RANDOM_REALS;           /* Compiler might mung the division itself */
  cd  = INIT_CD;
  cd /= RANDOM_REALS;
  cm  = INIT_CM;
  cm /= RANDOM_REALS;

  for (elt = 0;   elt < STATE_SIZE;   elt++) {
    s   = 0.0;
    bit = 1.0 / RANDOM_REALS;

    for (bit_number = 0;   bit_number < MANTISSA_SIZE;   bit_number++) {
      mm = (((ii * jj) % P) * kk) % P;
      ii = jj;
      jj = kk;
      kk = mm;
      ll = (53 * ll + 1) % Q;

      if (((ll * mm) % 64) >= 32) s += bit;
      bit += bit;
    }
    u[elt] = s;
  }
}

double ranNum(void)
{

/**
 | Returns a uniformly distributed pseudo random number in the range
 | [0.0, 1.0-2^(-24)] inclusive; there are 2^24 possible return values.
 | Side-effects on the non-local variables "u", "c", "ni" and "nj".
**/

  register double  uni;

  if (u[ni] < u[nj]) uni = u[ni] = u[ni] + (1.0 - u[nj]);
      else           uni = u[ni] = u[ni] - u[nj];

  if (!ni) ni = STATE_SIZE - 1; else ni--;
  if (!nj) nj = STATE_SIZE - 1; else nj--;

  c  =  c < cd  ?  c + (cm - cd)  :  c - cd;

  return  uni < c  ?  uni + (1.0 - c)  :  uni - c;
}
